from .scale import calculate_scaled_dimensions
from .scale import get_scale_mode
from collections.abc import MutableMapping
from persistent.dict import PersistentDict
from plone.scale.interfaces import IImageScaleFactory
from time import time
from ZODB.POSException import ConflictError
from zope.annotation import IAnnotations
from zope.interface import implementer
from zope.interface import Interface

import hashlib
import logging
import pprint


try:
    from plone.protect.utils import safeWrite
except ImportError:
    safeWrite = None

logger = logging.getLogger("plone.scale")
# Keep old scales around for this amount of milliseconds.
# This is one day:
KEEP_SCALE_MILLIS = 24 * 60 * 60 * 1000

# Number types are float and int, and on Python 2 also long.
number_types = [float]
number_types.extend((int,))
number_types = tuple(number_types)


class IImageScaleStorage(Interface):
    """This is an adapter for image content which can store, retrieve and
    generate image scale data. It provides a dictionary interface to
    existing image scales using the scale id as key. To find or create a
    scale based on its scaling parameters use the :meth:`scale` method."""

    def __init__(context, modified=None):
        """Adapt the given context item and optionally provide a callable
        to return a representation of the last modification date, which
        can be used to invalidate stored scale data on update."""

    def pre_scale(**parameters):
        """Pre-register a unique id for a scale, without actually scaling.

        In other words: no use of Pillow allowed.
        """

    def scale(**parameters):
        """Find image scale data for the given parameters or create it.

        We will look for an IImageScaleFactory for the context, and pass
        the parameters.  This factory is expected to return a tuple
        containing a representation of the actual image scale data (i.e.
        a string or file-like object) as well as the image's format and
        dimensions.  For convenience, this happens to match the return
        value of `scaleImage`, but makes it possible to use different
        storages, i.e. ZODB blobs"""

    def __getitem__(uid):
        """Find image scale data based on its uid."""

    def get_or_generate(uid):
        """Find image scale data based on its uid, or generate it.

        Only uids pre-generated by the pre_scale method should be accepted.

        Outdated scales may be returned: this method is called when requesting a scale
        on a cached, unique url.  An old cached page may still point to this.
        """


class ScalesDict(PersistentDict):
    def raise_conflict(self, saved, new):
        logger.info("Conflict")
        logger.debug("saved\n" + pprint.pformat(saved))
        logger.debug("new\n" + pprint.pformat(new))
        raise ConflictError

    def _p_resolveConflict(self, oldState, savedState, newState):
        logger.debug("Resolve conflict")
        old = oldState["data"]
        saved = savedState["data"]
        new = newState["data"]
        added = []
        modified = []
        deleted = []
        for key, value in new.items():
            if key not in old:
                added.append(key)
            elif value["modified"] != old[key]["modified"]:
                modified.append(key)
            # else:
            # unchanged
        for key in old:
            if key not in new:
                deleted.append(key)
        for key in deleted:
            if key in saved:
                if old[key]["modified"] == saved[key]["modified"]:
                    # unchanged by saved, deleted by new
                    logger.debug("deleted %s" % repr(key))
                    del saved[key]
                else:
                    # modified by saved, deleted by new
                    self.raise_conflict(saved[key], key)
        for key in added:
            if key in saved:
                # added by saved, added by new
                self.raise_conflict(saved[key], new[key])
            else:
                # not in saved, added by new
                logger.debug("added %s" % repr(key))
                saved[key] = new[key]
        for key in modified:
            if key not in saved:
                # deleted by saved, modified by new
                self.raise_conflict(key, new[key])
            elif saved[key]["modified"] != old[key]["modified"]:
                # modified by saved, modified by new
                self.raise_conflict(saved[key], new[key])
            else:
                # unchanged in saved, modified by new
                logger.debug("modified %s" % repr(key))
                saved[key] = new[key]
        return dict(data=saved)


@implementer(IImageScaleStorage)
class AnnotationStorage(MutableMapping):
    """An abstract storage for image scale data using annotations and
    implementing :class:`IImageScaleStorage`. Image data is stored as an
    annotation on the object container, i.e. the image. This is needed
    since not all images are themselves annotatable."""

    def __init__(self, context, modified=None):
        self.context = context
        self.modified = modified

    def _modified_since(self, since, offset=0):
        # offset gets subtracted from the main modified time: this allows to
        # keep scales for a bit longer if needed, even when the main image has
        # changed.
        if since is None:
            return False
        modified_time = self.modified_time
        if modified_time is None:
            return False
        # We expect a number, but in corner cases it can be
        # something else entirely.
        # https://github.com/plone/plone.scale/issues/12
        if not isinstance(modified_time, number_types):
            return False
        if not isinstance(since, number_types):
            return False
        modified_time = modified_time - offset
        return modified_time > since

    @property
    def modified_time(self):
        if self.modified is not None:
            return self.modified()
        else:
            return None

    def __repr__(self):
        name = self.__class__.__name__
        return f"<{name} context={self.context!r}>"

    __str__ = __repr__

    @property
    def storage(self):
        annotations = IAnnotations(self.context)
        if "plone.scale" not in annotations:
            annotations["plone.scale"] = ScalesDict()
            if safeWrite is not None:
                safeWrite(self.context)
        scales = annotations["plone.scale"]
        if not isinstance(scales, ScalesDict):
            # migrate from PersistentDict to ScalesDict
            new_scales = ScalesDict(scales)
            annotations["plone.scale"] = new_scales
            if safeWrite is not None:
                safeWrite(self.context)
            return new_scales
        return scales

    def hash(self, **parameters):
        return tuple(sorted(parameters.items()))

    def unhash(self, hash_key):
        return dict(hash_key)

    def get_info_by_hash(self, hash):
        for value in self.storage.values():
            if value["key"] == hash:
                return value

    def hash_key(self, **parameters):
        if "modified" in parameters:
            del parameters["modified"]
        key = self.hash(modified=self.modified_time, **parameters)
        fieldname = parameters.get("fieldname", "image")
        dimension = parameters.get("width", parameters.get("scale"))
        if dimension is None:
            dimension = 0
        hash_key = hashlib.md5(str(key).encode("utf-8")).hexdigest()
        # We return a uid that is recognizable when you inspect a url in html or
        # on the network tab: you immediately see for which field this is and what
        # the width is.  This helps during debugging/testing.
        return f"{fieldname}-{dimension}-{hash_key}"

    def pre_scale(self, **parameters):
        # This does *not* create a scale.
        # It only prepares info.
        logger.debug(f"Pre scale {parameters}")
        uid = self.hash_key(**parameters)
        # self.clear()
        # logger.debug(list(self.storage.keys()))
        info = self.get(uid)
        if info is not None and not self._modified_since(info["modified"]):
            logger.debug(f"Pre scale returns old {info}")
            return info

        scaling_factory = IImageScaleFactory(self.context, None)
        if scaling_factory is None:
            # There is nothing we can do.
            return
        fieldname = parameters.get("fieldname", None)
        value = scaling_factory.get_original_value(fieldname=fieldname)
        if value is None:
            # Either an empty value, or the field does not exist.
            return

        # There is no info, or it is outdated.  Recreate the scale info.
        # We need width and height for various reasons.
        # Start with a basis.
        width = parameters.get("width")
        height = parameters.get("height")
        orig_width, orig_height = value.getImageSize()
        mode = get_scale_mode(parameters.get("mode"), parameters.get("direction"))
        width, height = calculate_scaled_dimensions(
            orig_width, orig_height, width, height, mode
        )
        mimetype = value.contentType
        key = self.hash(**parameters)
        info = dict(
            uid=uid,
            key=key,
            modified=int(time() * 1000),
            mimetype=mimetype,
            data=None,
            width=width,
            height=height,
        )
        self.storage[uid] = info
        logger.debug(f"Pre scale returns new {info}")
        return info

    def generate_scale(self, uid=None, **parameters):
        logger.debug("Generating scale...")
        scaling_factory = IImageScaleFactory(self.context, None)
        if scaling_factory is None:
            # There is nothing we can do.
            return
        result = scaling_factory(**parameters)
        if result is None:
            return
        # storage will be modified:
        # good time to also cleanup
        fieldname = parameters.get("fieldname")
        self._cleanup(fieldname=fieldname)
        data, format_, dimensions = result
        width, height = dimensions
        if uid is None:
            uid = self.hash_key(**parameters)
        key = self.hash(**parameters)
        info = dict(
            uid=uid,
            data=data,
            width=width,
            height=height,
            mimetype=f"image/{format_.lower()}",
            key=key,
            modified=self.modified_time or int(time() * 1000),
        )
        if fieldname:
            info["fieldname"] = fieldname
        self.storage[uid] = info
        logger.debug(f"Generated scale: {info}")
        return info

    def scale(self, **parameters):
        logger.debug(f"scale called with {parameters}")
        uid = self.hash_key(**parameters)
        info = self.get(uid)
        if info is None:
            # Might be on old-style uuid4 scale
            key = self.hash(**parameters)
            info = self.get_info_by_hash(key)
        if (
            info is not None
            and info.get("data") is not None
            and not self._modified_since(info["modified"])
        ):
            logger.debug(f"scale found existing info {info}")
            return info
        return self.generate_scale(**parameters)

    def get_or_generate(self, name):
        logger.debug(f"get or generate {name}")
        info = self.get(name)
        if info is None:
            logger.debug(f"get or generate {name} not found")
            return
        if info is not None and info.get("data") is not None:
            # We could check 'self._modified_since(info["modified"])'.
            # But in fact we do not care if this scale is outdated.
            # A cached page may point to this, and the browser requests it now.
            logger.debug(f"get or generate {name} found {info}")
            return info
        # This scale has not been generated yet.
        # Get the parameters used when pre-registering this scale.
        parameters = self.unhash(info["key"])
        return self.generate_scale(uid=name, **parameters)

    def _cleanup(self, fieldname=None):
        storage = self.storage
        modified_time = self.modified_time
        if modified_time is None:
            return
        if not isinstance(modified_time, number_types):
            # https://github.com/plone/plone.scale/issues/12
            return
        for key, value in list(storage.items()):
            # remove info stored by tuple keys
            # before refactoring
            if isinstance(key, tuple):
                del self[key]
            if fieldname and "fieldname" in value and value["fieldname"] != fieldname:
                # Leave scales for other fieldnames alone.
                # self.modified may have nothing to do with that field.
                continue
            # clear cache from scales older than one day
            elif self._modified_since(value["modified"], offset=KEEP_SCALE_MILLIS):
                del self[key]

    def __getitem__(self, uid):
        return self.storage[uid]

    def __setitem__(self, id, scale):
        raise RuntimeError("New scales have to be created via scale()")

    def __delitem__(self, uid):
        try:
            del self.storage[uid]
        except KeyError:
            # This should not happen, but it apparently can happen in corner
            # cases.  See https://github.com/plone/plone.scale/issues/15
            logger.warning("Could not delete key %s from storage.", uid)

    def __iter__(self):
        return iter(self.storage)

    def __len__(self):
        return len(self.keys())

    def keys(self):
        return self.storage.keys()

    def has_key(self, uid):
        return uid in self.storage

    __contains__ = has_key

    def clear(self):
        self.storage.clear()



from operator import delitem
from operator import itemgetter
from operator import setitem
from plone.testing import zca
from unittest import TestCase
from zope.component import provideAdapter
from zope.interface import implementer

import zope.annotation.attribute
import zope.annotation.interfaces


_marker = object()


class DummyImage:
    contentType = "image/jpeg"

    def getImageSize(self):
        # width, height
        return 60, 40


@implementer(zope.annotation.interfaces.IAttributeAnnotatable)
class _DummyContext:
    pass


class AnnotationStorageTests(TestCase):
    layer = zca.UNIT_TESTING

    def _provide_dummy_scale_adapter(self, result=_marker):
        from plone.scale.interfaces import IImageScaleFactory
        from zope.component import adapter

        if result is _marker:
            result = DummyImage()
        factory = self.factory

        @implementer(IImageScaleFactory)
        @adapter(_DummyContext)
        class DummyISF:
            def __init__(self, context):
                self.context = context

            def __call__(self, **parameters):
                if result:
                    return factory()
                return None

            def get_original_value(self, fieldname=None):
                return result

        provideAdapter(DummyISF)

    @property
    def storage(self):
        from plone.scale.storage import AnnotationStorage

        provideAdapter(zope.annotation.attribute.AttributeAnnotations)
        storage = AnnotationStorage(_DummyContext())
        storage.modified = lambda: 42
        return storage

    def factory(self, **kw):
        return "some data", "png", (42, 23)

    def testInterface(self):
        from plone.scale.storage import IImageScaleStorage

        storage = self.storage
        self.assertTrue(IImageScaleStorage.providedBy(storage))

    def testScaleForNonExistingScaleWithCreation(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale = storage.scale(foo=23, bar=42)
        self.assertIn("uid", scale)
        self.assertIn("key", scale)
        self.assertEqual(scale["data"], "some data")
        self.assertEqual(scale["width"], 42)

        self.assertEqual(scale["height"], 23)
        self.assertEqual(scale["mimetype"], "image/png")

    def testScaleForNonExistingScaleWithoutCreation(self):
        self._provide_dummy_scale_adapter(result=None)
        storage = self.storage
        scale = storage.scale(foo=23, bar=42)
        self.assertEqual(scale, None)

    def testPreScaleNoWidthAndHeight(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        # It is actually pretty silly to not pass a width and height.
        # We get the original width and height then.
        scale = storage.pre_scale(foo=23, bar=42)
        self.assertIn("uid", scale)
        self.assertIn("key", scale)
        self.assertEqual(scale["data"], None)
        # We get the values from the DummyImage class.
        self.assertEqual(scale["width"], 60)
        self.assertEqual(scale["height"], 40)
        self.assertEqual(scale["mimetype"], "image/jpeg")

    def testPreScaleForNonExistingScale(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale = storage.pre_scale(width=50, height=80)
        self.assertIn("uid", scale)
        self.assertIn("key", scale)
        self.assertEqual(scale["data"], None)
        self.assertEqual(scale["width"], 50)
        # Note: the original image is 60x40.
        # By default we scale without cropping, so you do not get height 80.
        self.assertEqual(scale["height"], 33)
        self.assertEqual(scale["mimetype"], "image/jpeg")
        # Request the same pre scale.
        scale2 = storage.pre_scale(width=50, height=80)
        self.assertEqual(scale2["uid"], scale["uid"])
        self.assertEqual(scale2, scale)
        # This will really generate the scale.
        new_scale = storage.scale(width=50, height=80)
        self.assertEqual(new_scale["uid"], scale["uid"])
        self.assertIn("key", new_scale)
        self.assertEqual(new_scale["data"], "some data")
        # Our dummy adapter is silly and does not do anything with
        # the requested width and height.
        self.assertEqual(new_scale["width"], 42)
        self.assertEqual(new_scale["height"], 23)
        self.assertEqual(new_scale["mimetype"], "image/png")

        # Try cropping as well.
        scale = storage.pre_scale(width=50, height=80, mode="contain")
        self.assertIn("uid", scale)
        self.assertIn("key", scale)
        self.assertEqual(scale["data"], None)
        self.assertEqual(scale["width"], 50)
        self.assertEqual(scale["height"], 80)
        self.assertEqual(scale["mimetype"], "image/jpeg")

    def testPreScaleForNonExistingField(self):
        self._provide_dummy_scale_adapter(None)
        storage = self.storage
        scale = storage.pre_scale(width=50, height=80)
        self.assertIsNone(scale)
        # scale does the same.
        new_scale = storage.scale(width=50, height=80)
        self.assertIsNone(new_scale)

    def test_get_or_generate(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        self.assertIsNone(storage.get("random"))
        self.assertIsNone(storage.get_or_generate("random"))
        scale = storage.pre_scale(width=50, height=80)
        uid = scale["uid"]
        self.assertTrue(uid)
        # 'get' gets the pre generated placeholder info
        placeholder = storage.get(uid)
        self.assertEqual(placeholder["uid"], uid)
        self.assertIsNone(placeholder["data"])
        self.assertEqual(placeholder["width"], 50)
        self.assertEqual(placeholder["height"], 33)
        self.assertEqual(placeholder["mimetype"], "image/jpeg")
        # 'get_or_generate' gets the pre generated placeholder info
        # and generates the scale.
        real = storage.get_or_generate(uid)
        self.assertEqual(real["uid"], uid)
        self.assertEqual(real["data"], "some data")
        self.assertEqual(real["width"], 42)
        self.assertEqual(real["height"], 23)
        self.assertEqual(real["mimetype"], "image/png")

    def test_get_or_generate__stable_uid(self):
        # When get_or_generate actually generates the scale,
        # it should store it with the same uid that was used
        # to find the placeholder info, even if the field
        # has been modified or a modified time is not available
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale = storage.pre_scale(width=50, height=80)
        uid = scale["uid"]
        storage.modified = lambda: 100
        real = storage.get_or_generate(uid)
        self.assertEqual(real["uid"], uid)

    def testScaleForExistingScale(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale1 = storage.scale(foo=23, bar=42)
        scale2 = storage.scale(bar=42, foo=23)
        self.assertIs(scale1, scale2)

    def testScaleForSimilarScales(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale1 = storage.scale(foo=23, bar=42)
        scale2 = storage.scale(bar=42, foo=23, hurz="!")
        self.assertIsNot(scale1, scale2)

    def testGetItem(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale = storage.scale(foo=23, bar=42)
        uid = scale["uid"]
        scale = storage[uid]
        self.assertIn("uid", scale)
        self.assertIn("key", scale)
        self.assertEqual(scale["data"], "some data")
        self.assertEqual(scale["width"], 42)
        self.assertEqual(scale["height"], 23)
        self.assertEqual(scale["mimetype"], "image/png")

    def testGetUnknownItem(self):
        storage = self.storage
        self.assertRaises(KeyError, itemgetter("foo"), storage)

    def testSetItemNotAllowed(self):
        storage = self.storage
        self.assertRaises(RuntimeError, setitem, storage, "key", None)

    def testIterateWithoutAnnotations(self):
        storage = self.storage
        self.assertEqual(list(storage), [])

    def testIterate(self):
        storage = self.storage
        storage.storage.update(dict(one=None, two=None))
        generator = iter(storage)
        self.assertEqual(set(generator), {"one", "two"})

    def testKeys(self):
        storage = self.storage
        storage.storage.update(dict(one=None, two=None))
        self.assertEqual(set(storage.keys()), {"one", "two"})

    def testNegativeHasKey(self):
        storage = self.storage
        self.assertEqual("one" in storage, False)

    def testPositiveHasKey(self):
        storage = self.storage
        storage.storage.update(dict(one=None))
        self.assertEqual("one" in storage, True)

    def testDeleteNonExistingItem(self):
        storage = self.storage
        # This used to raise a KeyError, but sometimes the underlying storage
        # can get inconsistent, so it is nicer to accept it.
        # See https://github.com/plone/plone.scale/issues/15
        delitem(storage, "foo")

    def testDeleteRemovesItemAndIndex(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale = storage.scale(foo=23, bar=42)
        self.assertEqual(len(storage), 1)
        del storage[scale["uid"]]
        self.assertEqual(len(storage), 0)

    def test_modified_since(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        self.assertEqual(storage.modified(), 42)

        self.assertTrue(storage._modified_since(41))
        self.assertFalse(storage._modified_since(42))
        self.assertFalse(storage._modified_since(43))

        self.assertFalse(storage._modified_since(41, offset=1))
        self.assertTrue(storage._modified_since(40, offset=1))
        self.assertFalse(storage._modified_since(32, offset=10))
        self.assertTrue(storage._modified_since(32, offset=9))

    def testCleanUpOldItemsForSameParameters(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        orig_modified = storage.modified()
        storage.pre_scale(foo=23, bar=42)
        self.assertEqual(len(storage), 1)
        scale_old = storage.scale(foo=23, bar=42)
        self.assertEqual(len(storage), 1)
        storage.pre_scale(foo=23, bar=42)
        scale_old2 = storage.scale(foo=23, bar=42)
        self.assertEqual(len(storage), 1)
        self.assertEqual(scale_old, scale_old2)
        self.assertIn(scale_old["uid"], storage)
        next_modified = orig_modified + 10
        storage.modified = lambda: next_modified
        scale_new = storage.scale(foo=23, bar=42)
        self.assertEqual(len(storage), 2)
        self.assertIn(scale_new["uid"], storage)
        self.assertIn(scale_old["uid"], storage)
        next_modified = orig_modified + 24 * 60 * 60 * 1000 + 1
        storage.modified = lambda: next_modified
        scale_newer = storage.scale(foo=23, bar=42)
        self.assertIn(scale_newer["uid"], storage)
        self.assertIn(scale_new["uid"], storage)
        self.assertNotIn(scale_old["uid"], storage)
        del storage[scale_newer["uid"]]
        del storage[scale_new["uid"]]
        self.assertEqual(len(storage), 0)

    def testCleanUpOldItemsForDifferentParameters(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale_old = storage.scale(foo=23, bar=42)
        orig_modified = storage.modified()
        next_modified = orig_modified + 60000
        storage.modified = lambda: next_modified
        scale_new = storage.scale(foo=23, bar=50)
        self.assertEqual(len(storage), 2)
        self.assertIn(scale_new["uid"], storage)
        self.assertIn(scale_old["uid"], storage)

        # When modification time is older than a day, too old scales
        # get purged.
        next_modified = orig_modified + 24 * 60 * 60 * 1000 + 1
        storage.modified = lambda: next_modified
        scale_newer = storage.scale(foo=23, bar=70)

        self.assertIn(scale_newer["uid"], storage)
        self.assertIn(scale_new["uid"], storage)
        self.assertNotIn(scale_old["uid"], storage)

        next_modified = orig_modified + 24 * 60 * 60 * 1000 * 3
        storage.modified = lambda: next_modified
        scale_even_newer = storage.scale(foo=23, bar=42)
        self.assertIn(scale_even_newer["uid"], storage)
        self.assertNotIn(scale_newer["uid"], storage)
        self.assertNotIn(scale_new["uid"], storage)
        self.assertNotIn(scale_old["uid"], storage)
        del storage[scale_even_newer["uid"]]
        self.assertEqual(len(storage), 0)

    def testCleanUpOldItemsForDifferentFieldname(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        scale_image = storage.scale(fieldname="image", bar=42)
        next_modified = storage.modified() + 60000
        storage.modified = lambda: next_modified
        scale_leadimage_old = storage.scale(fieldname="leadimage", bar=50)
        self.assertEqual(len(storage), 2)
        self.assertIn(scale_leadimage_old["uid"], storage)
        self.assertIn(scale_image["uid"], storage)

        # When modification time is older than a day, too old scales
        # get purged.  But only for the current fieldname.
        next_modified = storage.modified() + 24 * 60 * 60 * 1000 + 1
        storage.modified = lambda: next_modified
        scale_leadimage_new = storage.scale(fieldname="leadimage", bar=70)

        self.assertIn(scale_leadimage_new["uid"], storage)
        self.assertNotIn(scale_leadimage_old["uid"], storage)
        self.assertIn(scale_image["uid"], storage)

        # If we manually call cleanup without a fieldname,
        # all items are checked.
        storage._cleanup()
        self.assertIn(scale_leadimage_new["uid"], storage)
        self.assertNotIn(scale_image["uid"], storage)
        del storage[scale_leadimage_new["uid"]]
        self.assertEqual(len(storage), 0)

    def testClear(self):
        self._provide_dummy_scale_adapter()
        storage = self.storage
        storage.scale(foo=23, bar=42)
        self.assertEqual(len(storage), 1)
        storage.clear()
        self.assertEqual(len(storage), 0)


def test_suite():
    from unittest import defaultTestLoader

    return defaultTestLoader.loadTestsFromName(__name__)



from io import BytesIO as StringIO
from plone.scale.scale import calculate_scaled_dimensions
from plone.scale.scale import scaleImage
from plone.scale.scale import scalePILImage
from plone.scale.tests import TEST_DATA_LOCATION
from unittest import TestCase

import functools
import os.path
import PIL.Image
import PIL.ImageDraw
import warnings


with open(os.path.join(TEST_DATA_LOCATION, "logo.png"), "rb") as fio:
    PNG = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "logo.gif"), "rb") as fio:
    GIF = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "logo.tiff"), "rb") as fio:
    TIFF = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "cmyk.jpg"), "rb") as fio:
    CMYK = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "profile.jpg"), "rb") as fio:
    PROFILE = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "profile.webp"), "rb") as fio:
    PROFILE_WEBP = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "animated.gif"), "rb") as fio:
    ANIGIF = fio.read()
with open(os.path.join(TEST_DATA_LOCATION, "animated2.gif"), "rb") as fio:
    ANIGIF2 = fio.read()


class ScalingTests(TestCase):
    def testNewSizeReturned(self):
        (imagedata, format, size) = scaleImage(PNG, 42, 51, "contain")
        input = StringIO(imagedata)
        image = PIL.Image.open(input)
        self.assertEqual(image.size, size)

    def testScaledImageKeepPNG(self):
        self.assertEqual(scaleImage(PNG, 84, 103, "contain")[1], "PNG")

    def testScaledImageKeepGIFto(self):
        self.assertEqual(scaleImage(GIF, 84, 103, "contain")[1], "PNG")

    def testScaledImageIsJpeg(self):
        self.assertEqual(scaleImage(TIFF, 84, 103, "contain")[1], "JPEG")

    def testScaledAnigifKeepGIF(self):
        self.assertEqual(scaleImage(ANIGIF, 84, 103, "contain")[1], "GIF")

    def testScaledAnigifKeepGIF2(self):
        self.assertEqual(scaleImage(ANIGIF2, 84, 103, "contain")[1], "GIF")

    def testAlphaForcesPNG(self):
        # first image without alpha
        src = PIL.Image.new("RGBA", (256, 256), (255, 255, 255, 255))
        for y in range(0, 256):
            for x in range(0, 256):
                src.putpixel((x, y), (x, y, 0, 255))
        result = StringIO()
        src.save(result, "TIFF")
        self.assertEqual(scaleImage(result, 84, 103, "contain")[1], "JPEG")
        # now with alpha
        src = PIL.Image.new("RGBA", (256, 256), (255, 255, 255, 128))
        result = StringIO()
        for y in range(0, 256):
            for x in range(0, 256):
                src.putpixel((x, y), (x, y, 0, x))
        src.save(result, "TIFF")
        self.assertEqual(scaleImage(result, 84, 103, "contain")[1], "PNG")

    def testScaledCMYKIsRGB(self):
        (imagedata, format, size) = scaleImage(CMYK, 42, 51, "contain")
        input = StringIO(imagedata)
        image = PIL.Image.open(input)
        self.assertEqual(image.mode, "RGB")

    def testScaledPngImageIsPng(self):
        self.assertEqual(scaleImage(PNG, 84, 103, "contain")[1], "PNG")

    def testScaledPreservesProfile(self):
        (imagedata, format, size) = scaleImage(PROFILE, 42, 51, "contain")
        input = StringIO(imagedata)
        image = PIL.Image.open(input)
        self.assertIsNotNone(image.info.get("icc_profile"))

    def testScaleWithFewColorsStaysColored(self):
        (imagedata, format, size) = scaleImage(PROFILE, 16, None, "contain")
        image = PIL.Image.open(StringIO(imagedata))
        self.assertEqual(max(image.size), 16)
        self.assertEqual(image.mode, "RGB")
        self.assertEqual(image.format, "JPEG")

    def testScaledWebp(self):
        (imagedata, format, size) = scaleImage(PROFILE_WEBP, 120, 120)
        self.assertEqual(format, "WEBP")
        self.assertEqual(size, (120, 120))
        self.assertTrue(len(imagedata) < len(PROFILE_WEBP))
        input = StringIO(imagedata)
        image = PIL.Image.open(input)
        self.assertIsNotNone(image.info.get("icc_profile"))

    def testAutomaticGreyscale(self):
        src = PIL.Image.new("RGB", (256, 256), (255, 255, 255))
        draw = PIL.ImageDraw.Draw(src)
        for i in range(0, 256):
            draw.line(((0, i), (256, i)), fill=(i, i, i))
        result = StringIO()
        src.save(result, "JPEG")
        (imagedata, format, size) = scaleImage(result, 200, None, "contain")
        image = PIL.Image.open(StringIO(imagedata))
        self.assertEqual(max(image.size), 200)
        self.assertEqual(image.mode, "L")
        self.assertEqual(image.format, "JPEG")

    def testAutomaticPalette(self):
        # get a JPEG with more than 256 colors
        jpeg = PIL.Image.open(StringIO(PROFILE))
        self.assertEqual(jpeg.mode, "RGB")
        self.assertEqual(jpeg.format, "JPEG")
        self.assertIsNone(jpeg.getcolors(maxcolors=256))
        # convert to PNG
        dst = StringIO()
        jpeg.save(dst, "PNG")
        dst.seek(0)
        png = PIL.Image.open(dst)
        self.assertEqual(png.mode, "RGB")
        self.assertEqual(png.format, "PNG")
        self.assertIsNone(png.getcolors(maxcolors=256))
        # scale it to a size where we get less than 256 colors
        (imagedata, format, size) = scaleImage(dst.getvalue(), 24, None, "contain")
        image = PIL.Image.open(StringIO(imagedata))
        # we should now have an image in palette mode
        self.assertEqual(image.mode, "P")
        self.assertEqual(image.format, "PNG")

    def testSameSizeDownScale(self):
        self.assertEqual(scaleImage(PNG, 84, 103, "contain")[2], (84, 103))

    def testHalfSizeDownScale(self):
        self.assertEqual(scaleImage(PNG, 42, 51, "contain")[2], (42, 51))

    def testScaleWithCropDownScale(self):
        self.assertEqual(scaleImage(PNG, 20, 51, "contain")[2], (20, 51))

    def testNoStretchingDownScale(self):
        self.assertEqual(scaleImage(PNG, 200, 103, "contain")[2], (200, 103))

    def testHugeScale(self):
        # The image will be cropped, but not scaled.
        # If such a ridiculous height is given, we only look at the width.
        self.assertEqual(scaleImage(PNG, 400, 99999, "contain")[2], (400, 400))

    def testZeroHeightScale(self):
        # In this case we only look at the width.
        self.assertEqual(scaleImage(PNG, 400, 0, "contain")[2], (400, 400))

    def testNegativeHeightScale(self):
        # In this case we only look at the width.
        self.assertEqual(scaleImage(PNG, 400, -1, "contain")[2], (400, 400))

    def testCropPreWideScaleUnspecifiedHeight(self):
        image = scaleImage(PNG, 400, None, "contain")
        self.assertEqual(image[2], (400, 400))

    def testCropPreWideScale(self):
        image = scaleImage(PNG, 400, 100, "contain")
        self.assertEqual(image[2], (400, 100))

    def testCropPreTallScaleUnspecifiedWidth(self):
        image = scaleImage(PNG, None, 400, "contain")
        self.assertEqual(image[2], (326, 400))

    def testCropPreTallScale(self):
        image = scaleImage(PNG, 100, 400, "contain")
        self.assertEqual(image[2], (100, 400))

    def testRestrictWidthOnlyDownScaleNone(self):
        self.assertEqual(scaleImage(PNG, 42, None, "contain")[2], (42, 42))

    def testRestrictWidthOnlyDownScaleZero(self):
        self.assertEqual(scaleImage(PNG, 42, 0, "contain")[2], (42, 42))

    def testRestrictHeightOnlyDownScaleNone(self):
        self.assertEqual(scaleImage(PNG, None, 51, "contain")[2], (42, 51))

    def testRestrictHeightOnlyDownScaleZero(self):
        self.assertEqual(scaleImage(PNG, 0, 51, "contain")[2], (42, 51))

    def testSameSizeUpScale(self):
        self.assertEqual(scaleImage(PNG, 84, 103, "cover")[2], (84, 103))

    def testDoubleSizeUpScale(self):
        self.assertEqual(scaleImage(PNG, 168, 206, "cover")[2], (168, 206))

    def testHalfSizeUpScale(self):
        self.assertEqual(scaleImage(PNG, 42, 51, "cover")[2], (42, 51))

    def testNoStretchingUpScale(self):
        self.assertEqual(scaleImage(PNG, 200, 103, "cover")[2], (84, 103))

    def testRestrictWidthOnlyUpScaleNone(self):
        self.assertEqual(scaleImage(PNG, 42, None, "cover")[2], (42, 52))

    def testRestrictWidthOnlyUpScaleZero(self):
        self.assertEqual(scaleImage(PNG, 42, 0, "cover")[2], (42, 52))

    def testRestrictHeightOnlyUpScaleNone(self):
        self.assertEqual(scaleImage(PNG, None, 51, "cover")[2], (42, 51))

    def testRestrictHeightOnlyUpScaleZero(self):
        self.assertEqual(scaleImage(PNG, 0, 51, "cover")[2], (42, 51))

    def testNoRestrictionsNone(self):
        self.assertRaises(ValueError, scaleImage, PNG, None, None)

    def testNoRestrictionsZero(self):
        self.assertRaises(ValueError, scaleImage, PNG, 0, 0)

    def testKeepAspectRatio(self):
        self.assertEqual(scaleImage(PNG, 80, 80, "scale")[2], (65, 80))

    def testThumbnailHeightNone(self):
        self.assertEqual(scaleImage(PNG, 42, None, "scale")[2], (42, 51))

    def testThumbnailWidthNone(self):
        self.assertEqual(scaleImage(PNG, None, 51, "scale")[2], (41, 51))

    def testQuality(self):
        img1 = scaleImage(CMYK, 84, 103)[0]
        img2 = scaleImage(CMYK, 84, 103, quality=50)[0]
        img3 = scaleImage(CMYK, 84, 103, quality=20)[0]
        self.assertNotEqual(img1, img2)
        self.assertNotEqual(img1, img3)
        self.assertTrue(len(img1) > len(img2) > len(img3))

    def testResultBuffer(self):
        img1 = scaleImage(PNG, 84, 103)[0]
        result = StringIO()
        img2 = scaleImage(PNG, 84, 103, result=result)[0]
        self.assertEqual(result, img2)  # the return value _is_ the buffer
        self.assertEqual(result.getvalue(), img1)  # but with the same value

    def testAlternativeSpellings(self):
        """Test alternative and deprecated mode spellings and the old
        ``direction`` arguments instead of ``mode``.
        """

        # Test mode contain.  This can do cropping.
        # scale-crop-to-fit
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 5, direction="scale-crop-to-fit")
        self.assertEqual(img_scaled.size, (10, 5))
        # down
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 5, direction="down")
        self.assertEqual(img_scaled.size, (10, 5))

        # Test mode cover
        # scale-crop-to-fill
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 40, 30, direction="scale-crop-to-fill")
        self.assertEqual(img_scaled.size, (30, 30))
        # up
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 40, 30, direction="up")
        self.assertEqual(img_scaled.size, (30, 30))

        # Test mode scale
        # keep A
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 20, 10, direction="keep")
        self.assertEqual(img_scaled.size, (10, 10))
        # keep B
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 40, 80, direction="keep")
        self.assertEqual(img_scaled.size, (20, 20))
        # thumbnail A
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 20, 10, direction="thumbnail")
        self.assertEqual(img_scaled.size, (10, 10))
        # thumbnail B
        img = PIL.Image.new("RGB", (20, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 40, 80, direction="thumbnail")
        self.assertEqual(img_scaled.size, (20, 20))

    def testModes(self):
        """Test modes to actually behavie like documented."""
        # Mode contain
        # v
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="contain")
        self.assertEqual(img_scaled.size, (10, 10))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="contain")
        self.assertEqual(img_scaled.size, (10, 10))
        # ^
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="contain")
        self.assertEqual(img_scaled.size, (60, 60))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="contain")
        self.assertEqual(img_scaled.size, (60, 60))

        # Mode cover
        # v
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="cover")
        self.assertEqual(img_scaled.size, (5, 10))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="cover")
        self.assertEqual(img_scaled.size, (10, 5))
        # ^
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="cover")
        self.assertEqual(img_scaled.size, (30, 60))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="cover")
        self.assertEqual(img_scaled.size, (60, 30))

        # Mode scale
        # v
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="scale")
        self.assertEqual(img_scaled.size, (5, 10))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 10, 10, mode="scale")
        self.assertEqual(img_scaled.size, (10, 5))
        # ^
        # A
        img = PIL.Image.new("RGB", (20, 40), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="scale")
        self.assertEqual(img_scaled.size, (20, 40))
        # B
        img = PIL.Image.new("RGB", (40, 20), (0, 0, 0))
        img_scaled = scalePILImage(img, 60, 60, mode="scale")
        self.assertEqual(img_scaled.size, (40, 20))

    def testDeprecations(self):
        import plone.scale.scale

        # clear warnings registry, so the test actually sees the warning
        plone.scale.scale.__warningregistry__.clear()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            scaleImage(PNG, 16, 16, direction="keep")
            self.assertEqual(len(w), 1)
            self.assertIs(w[0].category, DeprecationWarning)
            self.assertIn("The 'direction' option is deprecated", str(w[0].message))

    def testDeprecationsAni(self):
        import plone.scale.scale

        # clear warnings registry, so the test actually sees the warning
        plone.scale.scale.__warningregistry__.clear()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            scaleImage(ANIGIF, 16, 16, direction="keep")
            self.assertEqual(len(w), 6)
            for item in w:
                self.assertIs(item.category, DeprecationWarning)
                self.assertIn("The 'direction' option is deprecated", str(item.message))

    def test_calculate_scaled_dimensions_contain(self):
        """Test the calculate_scaled_dimensions function with mode "contain".

        You pass it:

            original_width, original_height, width, height

        Plus an optional mode, by default "scale"`.
        Alternative spellings: `scale-crop-to-fit`, `down`.
        """
        calc = functools.partial(calculate_scaled_dimensions, mode="contain")
        self.assertEqual(calc(1, 1, 1, 1), (1, 1))
        self.assertEqual(calc(10, 10, 1, 1), (1, 1))
        self.assertEqual(calc(1, 1, 10, 10), (10, 10))
        self.assertEqual(calc(10, 20, 10, 10), (10, 10))
        # Try the new preview scale.
        # This is defined as width 400 and a very large height.
        # That does not work at all for cropping.
        self.assertEqual(calc(10, 20, 400, 65536), (400, 400))
        self.assertEqual(calc(600, 300, 400, 65536), (400, 400))
        self.assertEqual(calc(600, 1200, 400, 65536), (400, 400))

    def test_calculate_scaled_dimensions_cover(self):
        """Test calculate_scaled_dimensions function with mode "cover".

        Alternative spellings: `scale-crop-to-fill`, `up`.
        Despite what you may think, this does not crop.
        """
        calc = functools.partial(calculate_scaled_dimensions, mode="cover")
        self.assertEqual(calc(1, 1, 1, 1), (1, 1))
        self.assertEqual(calc(10, 10, 1, 1), (1, 1))
        # Mode "cover" scales up:
        self.assertEqual(calc(1, 1, 10, 10), (10, 10))
        # If any cropping would happen, the next answer would be (10, 10):
        self.assertEqual(calc(10, 20, 10, 10), (5, 10))
        # Try the new preview scale:
        self.assertEqual(calc(10, 20, 400, 65536), (400, 800))
        self.assertEqual(calc(600, 300, 400, 65536), (400, 200))
        self.assertEqual(calc(600, 1200, 400, 65536), (400, 800))

    def test_calculate_scaled_dimensions_scale(self):
        """Test calculate_scaled_dimensions function with mode "scale".

        Alternative spellings: `keep`, `thumbnail`.
        "scale" is the default
        """
        # calc = functools.partial(calculate_scaled_dimensions, mode="scale")
        calc = calculate_scaled_dimensions
        self.assertEqual(calc(1, 1, 1, 1), (1, 1))
        self.assertEqual(calc(10, 10, 1, 1), (1, 1))
        # Mode "scale" only scales down, not up:
        self.assertEqual(calc(1, 1, 10, 10), (1, 1))
        self.assertEqual(calc(10, 20, 10, 10), (5, 10))
        # Try the new preview scale:
        self.assertEqual(calc(10, 20, 400, 65536), (10, 20))
        self.assertEqual(calc(600, 300, 400, 65536), (400, 200))
        self.assertEqual(calc(600, 1200, 400, 65536), (400, 800))

    def testAnimatedGifContainsAllFrames(self):
        image = scaleImage(ANIGIF, 84, 103, "contain")[0]
        with PIL.Image.open(StringIO(image)) as img:
            frames = [frame for frame in PIL.ImageSequence.Iterator(img)]
            self.assertEqual(len(frames), 6)
            for frame in frames:
                self.assertEqual(frame.width, 84)
                self.assertEqual(frame.height, 103)

    def testAnimatedGifContainsAllFrames2(self):
        image = scaleImage(ANIGIF2, 84, 103, "contain")[0]
        with PIL.Image.open(StringIO(image)) as img:
            frames = [frame for frame in PIL.ImageSequence.Iterator(img)]
            self.assertEqual(len(frames), 35)
            for frame in frames:
                self.assertEqual(frame.width, 84)
                self.assertEqual(frame.height, 103)


def test_suite():
    from unittest import defaultTestLoader

    return defaultTestLoader.loadTestsFromName(__name__)



from zope.interface import Interface


class IScaledImageQuality(Interface):
    """Marker interface for utility query.

    This can be used to define a property "scaled image quality" in the site's image
    handling settings.
    """


class IImageScaleFactory(Interface):
    """Creates a scale"""

    def __call__(
        fieldname=None, mode="scale", height=None, width=None, scale=None, **parameters
    ):
        """Interface defining an actual scaling operation.

        Arguments are:

        ``context``
            some object with images on

        ``fieldname``
            name of the field to scale

        ``mode``
            See ``scalePILImage`` for the values that should be accepted.
            This used to be called "direction", which may still be accepted,
            but is deprecated.

        ``width`` and ``height``
            target size

        ``scale``
            name of the current scale, if there is one. Can be used to retrieve
             additional information such as cropping boxes.

        ``**parameters``
            is a dict with optional additional expected keyword arguments

        Expected to return a triple of ``value, format, dimensions``
        or ``None`` on failure.

        ``value``
            is expected to be an storable value

        ``format``
            is the minor part of the ``image`` mimetype

        ``dimensions``
            is a tuple (width, height)
        """

    def get_original_value(fieldname=None):
        """Get the image value.

        In most cases this will be a NamedBlobImage field.
        Should accept an optional fieldname keyword argument.
        If not passed, and there is no self.fieldname set,
        you can try to get it in a different way.
        """



from io import BytesIO as StringIO

import math
import PIL.Image
import PIL.ImageFile
import PIL.ImageSequence
import sys
import warnings


try:
    # Pillow 9.1.0+
    LANCZOS = PIL.Image.Resampling.LANCZOS
except AttributeError:
    LANCZOS = PIL.Image.ANTIALIAS

# When height is higher than this we do not limit the height, but only the width.
# Otherwise cropping does not make sense, and in a Pillow you may get an error.
# In a Pillow traceback I saw 65500 as maximum.
# Several Plone scale definitions have 65536 (2**16).
# So pick a number slightly lower for good measure.
# A different idea was to use -1 here, so we support this:
# a height of 0 or less is ignored.
MAX_HEIGHT = 65000


def none_as_int(the_int):
    """For python 3 compatibility, to make int vs. none comparison possible
    without changing the algorithms below.

    This should mimic python2 behaviour."""
    if the_int is None:
        return -sys.maxsize
    return the_int


# Set a larger buffer size. This fixes problems with jpeg decoding.
# See http://mail.python.org/pipermail/image-sig/1999-August/000816.html for
# details.
PIL.ImageFile.MAXBLOCK = 1000000

MAX_PIXELS = 8192 * 8192


def scaleImage(
    image,
    width=None,
    height=None,
    mode="scale",
    quality=88,
    result=None,
    direction=None,
):
    """Scale the given image data to another size and return the result
    as a string or optionally write in to the file-like `result` object.

    The `image` parameter can either be the raw image data (ie a `str`
    instance) or an open file.

    The `quality` parameter can be used to set the quality of the
    resulting image scales.

    The return value is a tuple with the new image, the image format and
    a size-tuple.  Optionally a file-like object can be given as the
    `result` parameter, in which the generated image scale will be stored.

    The `width`, `height`, `mode` parameters will be passed to
    :meth:`scalePILImage`, which performs the actual scaling.

    The generated image is a JPEG image, unless the original is a WEBP, PNG
    or GIF image. This is needed to make sure alpha channel information is
    not lost, which JPEG does not support.
    """
    if isinstance(image, (bytes, str)):
        image = StringIO(image)

    animated_kwargs = {}
    with PIL.Image.open(image) as img:
        icc_profile = img.info.get("icc_profile")
        # When we create a new image during scaling we loose the format
        # information, so remember it here.
        format_ = img.format
        if format_ == "GIF":
            # Attempt to process multiple frames, to support animated GIFs
            append_images = []
            for frame in PIL.ImageSequence.Iterator(img):
                # We ignore the returned format_ as it won't get optimized
                # in case of a GIF. This ensures that the format remains
                # constant across all frames.
                scaled_frame, _dummy_format_ = scaleSingleFrame(
                    frame,
                    width=width,
                    height=height,
                    mode=mode,
                    format_=format_,
                    quality=quality,
                    direction=direction,
                )
                append_images.append(scaled_frame)

            # The first image is the basis for save
            # All other images than the first will be added as a save parameter
            image = append_images.pop(0)
            if len(append_images) > 0:
                # Saving as a multi page image
                animated_kwargs["save_all"] = True
                animated_kwargs["append_images"] = append_images
            else:
                # GIF scaled looks better if we have 8-bit alpha and no palette,
                # but it only works for single frame, so don't do this for animated GIFs.
                format_ = "PNG"

        else:
            # All other formats only process a single frame
            if format_ not in ("PNG", "GIF", "WEBP"):
                # Always generate JPEG, except if format is WEBP, PNG or GIF.
                format_ = "JPEG"
            image, format_ = scaleSingleFrame(
                img,
                width=width,
                height=height,
                mode=mode,
                format_=format_,
                quality=quality,
                direction=direction,
            )

    new_result = False
    if result is None:
        result = StringIO()
        new_result = True

    image.save(
        result,
        format_,
        quality=quality,
        optimize=True,
        progressive=True,
        icc_profile=icc_profile,
        **animated_kwargs,
    )

    if new_result:
        result = result.getvalue()
    else:
        result.seek(0)

    return result, format_, image.size


def scaleSingleFrame(
    image,
    width,
    height,
    mode,
    format_,
    quality,
    direction,
):
    image = scalePILImage(image, width, height, mode, direction=direction)

    # convert to simpler mode if possible
    colors = image.getcolors(maxcolors=256)
    if image.mode not in ("P", "L") and colors:
        if format_ == "JPEG":
            # check if it's all grey
            if all(rgb[0] == rgb[1] == rgb[2] for c, rgb in colors):
                image = image.convert("L")
        elif format_ in ("PNG", "GIF"):
            image = image.convert("P")

    if image.mode == "RGBA" and format_ == "JPEG":
        extrema = dict(zip(image.getbands(), image.getextrema()))
        if extrema.get("A") == (255, 255):
            # no alpha used, just change the mode, which causes the alpha band
            # to be dropped on save
            image = image.convert("RGB")
        else:
            # switch to PNG, which supports alpha
            format_ = "PNG"

    return image, format_


def _scale_thumbnail(image, width=None, height=None):
    """Scale with method "thumbnail".

    Aspect Ratio is kept. Resulting image has to fit in the given box.
    If target aspect ratio is different, either width or height is smaller
    than the given target width or height. No cropping!
    """
    dimensions = _calculate_all_dimensions(
        image.size[0], image.size[1], width, height, "scale"
    )

    if (dimensions.target_width * dimensions.target_height) > MAX_PIXELS:
        # The new image would be excessively large and eat up all memory while
        # scaling, so return the potentially pre cropped image
        return image

    image.draft(image.mode, (dimensions.target_width, dimensions.target_height))
    image = image.resize((dimensions.target_width, dimensions.target_height), LANCZOS)
    return image


def get_scale_mode(mode, direction=None):
    if direction is not None:
        warnings.warn(
            "The 'direction' option is deprecated, use 'mode' instead.",
            DeprecationWarning,
        )
        mode = direction

    if mode in ("scale", "keep", "thumbnail", None):
        return "scale"
    if mode in ("contain", "scale-crop-to-fit", "down"):
        return "contain"
    if mode in ("cover", "scale-crop-to-fill", "up"):
        return "cover"

    return mode


class ScaledDimensions:
    def __init__(self, original_width=0, original_height=0):
        self.final_width = self.target_width = original_width
        self.final_height = self.target_height = original_height
        self.factor_width = self.factor_height = 1.0
        self.post_scale_crop = False
        self.pre_scale_crop = False


def _calculate_all_dimensions(
    original_width, original_height, width, height, mode="scale"
):
    """Calculate all dimensions we need for scaling.

    final_width and final_height are the dimensions of the resulting image and
    are always present.

    The other values are required for cropping and scaling.
    """
    if height is not None and (height >= MAX_HEIGHT or height <= 0):
        height = None

    if mode not in ("contain", "cover", "scale"):
        raise ValueError("Unknown scale mode '%s'" % mode)

    dimensions = ScaledDimensions(
        original_width=original_width,
        original_height=original_height,
    )
    if width is None and height is None:
        return dimensions

    if mode == "scale":
        # calculate missing sizes
        if width is None:
            width = float(original_width) / float(original_height) * height
        elif height is None:
            height = float(original_height) / float(original_width) * width

        # keep aspect ratio of original
        target_width = original_width
        target_height = original_height
        if target_width > width:
            target_height = int(max(target_height * width / target_width, 1))
            target_width = int(width)
        if target_height > height:
            target_width = int(max(target_width * height / target_height, 1))
            target_height = int(height)

        dimensions.target_width = target_width
        dimensions.target_height = target_height

        if (target_width * target_height) > MAX_PIXELS:
            # The new image would be excessively large and eat up all memory while
            # scaling, so return the dimensions of the potentially cropped image
            return dimensions

        dimensions.final_width = dimensions.target_width
        dimensions.final_height = dimensions.target_height
        return dimensions

    # now for 'cover' and 'contain' scaling
    if mode == "contain" and height is None:
        # For cropping we need a height.
        height = width

    # Determine scale factors needed
    factor_height = factor_width = None
    if height is not None:
        factor_height = float(height) / float(original_height)
    if width is not None:
        factor_width = float(width) / float(original_width)

    dimensions.factor_width = factor_width
    dimensions.factor_height = factor_height
    dimensions.final_width = width
    dimensions.final_height = height

    if factor_height == factor_width:
        # The original already has the right aspect ratio
        return dimensions

    # figure out which axis to scale. One of the factors can still be None!
    use_height = none_as_int(factor_width) > none_as_int(factor_height)
    if mode == "cover":  # for 'cover': invert
        use_height = not use_height

    # keep aspect ratio
    if height is None or (use_height and width is not None):
        target_width = width
        target_height = int(round(original_height * factor_width))

    if width is None or (height is not None and not use_height):
        target_width = int(round(original_width * factor_height))
        target_height = height

    # determine whether we need to crop before scaling
    pre_scale_crop = (width is not None and target_width > width) or (
        height is not None and target_height > height
    )
    dimensions.pre_scale_crop = pre_scale_crop

    if pre_scale_crop:
        # crop image before scaling to avoid excessive memory use
        if use_height:
            left = 0
            right = original_width
            top = int(math.floor(((target_height - height) / 2.0) / factor_width))
            bottom = int(
                math.ceil((((target_height - height) / 2.0) + height) / factor_width)
            )
            pre_scale_crop_height = bottom - top
            # set new height in case we abort
            dimensions.final_height = pre_scale_crop_height
            # calculate new scale target_height from cropped height
            target_height = int(round(pre_scale_crop_height * factor_width))
        else:
            left = int(math.floor(((target_width - width) / 2.0) / factor_height))
            right = int(
                math.ceil((((target_width - width) / 2.0) + width) / factor_height)
            )
            top = 0
            bottom = original_height
            pre_scale_crop_width = right - left
            # set new width in case we abort
            dimensions.final_width = pre_scale_crop_width
            # calculate new scale target_width from cropped width
            target_width = int(round(pre_scale_crop_width * factor_height))
        dimensions.pre_scale_crop = (left, top, right, bottom)

    dimensions.target_width = target_width
    dimensions.target_height = target_height

    if (target_width * target_height) > MAX_PIXELS:
        # The new image would be excessively large and eat up all memory while
        # scaling, so return the dimensions of the potentially cropped image
        return dimensions

    dimensions.final_width = target_width
    dimensions.final_height = target_height

    # determine whether we have to crop after scaling due to rounding
    post_scale_crop = (width is not None and target_width > width) or (
        height is not None and target_height > height
    )
    dimensions.post_scale_crop = post_scale_crop

    if post_scale_crop:
        if use_height:
            left = 0
            right = target_width
            top = int((target_height - height) / 2.0)
            bottom = top + height
            dimensions.final_height = bottom - top
        else:
            left = int((target_width - width) / 2.0)
            right = left + width
            top = 0
            bottom = target_height
            dimensions.final_width = right - left
        dimensions.post_scale_crop = (left, top, right, bottom)

    return dimensions


def calculate_scaled_dimensions(
    original_width, original_height, width, height, mode="scale"
):
    """Calculate the scaled image dimensions from the originals using the
    same logic as scalePILImage"""
    dimensions = _calculate_all_dimensions(
        original_width, original_height, width, height, mode
    )

    return (dimensions.final_width, dimensions.final_height)


def scalePILImage(image, width=None, height=None, mode="scale", direction=None):
    """Scale a PIL image to another size.

    This is all about scaling for the display in a web browser.

    Either width or height - or both - must be given.

    Three different scaling options are supported via `mode` and correspond to
    the CSS background-size values
    (see https://developer.mozilla.org/en-US/docs/Web/CSS/background-size):

    `contain`
        Alternative spellings: `scale-crop-to-fit`, `down`.
        Starts by scaling the relatively smallest dimension to the required
        size and crops the other dimension if needed.

    `cover`
        Alternative spellings: `scale-crop-to-fill`, `up`.
        Scales the relatively largest dimension up to the required size.
        Despite the alternative spelling, I see no cropping happening.

    `scale`
        Alternative spellings: `keep`, `thumbnail`.
        Scales to the requested dimensions without cropping. The resulting
        image may have a different size than requested. This option
        requires both width and height to be specified.
        Does not scale up.

    The `image` parameter must be an instance of the `PIL.Image` class.

    The return value the scaled image in the form of another instance of
    `PIL.Image`.
    """
    # convert zero to None, same semantics: calculate this scale
    if width == 0:
        width = None
    if height == 0:
        height = None
    if width is None and height is None:
        raise ValueError("Either width or height need to be given")

    mode = get_scale_mode(mode, direction)

    if image.mode == "1":
        # Convert black&white to grayscale
        image = image.convert("L")
    elif image.mode == "P":
        # If palette is grayscale, convert to gray+alpha
        # Else convert palette based images to 3x8bit+alpha
        palette = image.getpalette()
        if palette[0::3] == palette[1::3] == palette[2::3]:
            image = image.convert("LA")
        else:
            image = image.convert("RGBA")
    elif image.mode == "CMYK":
        # Convert CMYK to RGB, allowing for web previews of print images
        image = image.convert("RGB")

    # for scale we're done:
    if mode == "scale":
        return _scale_thumbnail(image, width, height)

    dimensions = _calculate_all_dimensions(
        image.size[0], image.size[1], width, height, mode
    )

    if dimensions.factor_height == dimensions.factor_width:
        # The original already has the right aspect ratio, so we only need
        # to scale.
        if mode == "contain":
            image.thumbnail((dimensions.final_width, dimensions.final_height), LANCZOS)
            return image
        return image.resize((dimensions.final_width, dimensions.final_height), LANCZOS)

    if dimensions.pre_scale_crop:
        # crop image before scaling to avoid excessive memory use
        # in case the intermediate result would be very tall or wide
        image = image.crop(dimensions.pre_scale_crop)

    if (dimensions.target_width * dimensions.target_height) > MAX_PIXELS:
        # The new image would be excessively large and eat up all memory while
        # scaling, so return the potentially pre cropped image
        return image

    image.draft(image.mode, (dimensions.target_width, dimensions.target_height))
    image = image.resize((dimensions.target_width, dimensions.target_height), LANCZOS)

    if dimensions.post_scale_crop:
        # crop off remains due to rounding before scaling
        image = image.crop(dimensions.post_scale_crop)

    return image



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "4.1.1"

long_description = (
    f"{Path('README.rst').read_text()}\n{Path('CHANGES.rst').read_text()}"
)

STORAGE_REQUIREMENTS = [
    "ZODB",
    "zope.annotation",
    "zope.interface",
    "persistent",
]

TEST_REQUIREMENTS = [
    "zope.component",
    "plone.testing[test]",
]

setup(
    name="plone.scale",
    version=version,
    description="Image scaling",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="image scaling",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.scale",
    license="BSD",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    test_suite="plone.scale",
    python_requires=">=3.8",
    install_requires=[
        "Pillow",
        "setuptools",
        "zope.annotation",
    ],
    extras_require=dict(
        storage=STORAGE_REQUIREMENTS,
        test=STORAGE_REQUIREMENTS + TEST_REQUIREMENTS,
    ),
)
